#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate.tests.utility as utility
import opengate_core as g4

if __name__ == "__main__":
    a = "titi"
    s = g4.G4String(a)
    print("G4String =", s, type(s))
    assert s == "titi"
    assert type(s) == g4.G4String

    t = g4.G4String(s)
    print("G4String =", t, type(t))
    assert t == "titi"

    t = "toto"
    print(f"s={s} t={t} ", type(t))

    assert s == "titi"
    assert t == "toto"

    utility.test_ok(True)
