#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.tests.src import test022_half_life


def main():
    test022_half_life.test022_half_life(n_threads=3)


if __name__ == "__main__":
    main()
