# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
from datetime import datetime
import os
import sys
from pathlib import Path

# this variable is used by opengate_core to detect that the import is triggered on ReadTheDocs
# and no Geant4 data should be downloaded
os.environ["GATEONRTD"] = "1"

# sys.path.append(str(Path(__file__).resolve().parents[1]))
# print("DEBUG: sys.path = ", sys.path)
sys.path.append(str(Path("..", "..").resolve()))
autodoc_mock_imports = [
    "opengate_core",
    # "colored",
    # "gatetools",
    # "click",
    # "python-box",
    # "anytree",
    # "numpy",
    # "itk",
    # "uproot",
    # "scipy",
    # "matplotlib",
    # "GitPython",
    # "colorlog",
    # "meshio",
    # "radioactivedecay",
    # "jsonpickle",
    # "pandas",
    # "requests",
    # "PyYAML",
]

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

# -- Project information -----------------------------------------------------

project = "OPENGATE"
now = datetime.now()
copyright = (
    str(now.year) + ", OpenGATE Collaboration. Updated on " + now.strftime("%B %d, %Y")
)
author = "OpenGate Collaboration"

# The short X.Y version
version = ""
# The full version, including alpha/beta/rc tags
release = ""

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.linkcode",
    "sphinx_copybutton",
    # "autoapi.extension",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = {
    ".rst": "restructuredtext",
    # ".txt": "markdown",
    ".md": "markdown",
}

# The master toctree document.
master_doc = "index"

# Use number references for figures
numfig = False
# numfig_secnum_depth = 1
html_secnumber_suffix = ""

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None

# autodoc_default_options = {
#     'members': True,
#     'member-order': 'bysource',
#     'special-members': '__init__',
#     'undoc-members': False,
#     'exclude-members': '__weakref__'
# }
add_module_names = False


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'sphinx_rtd_theme'
html_theme = "pydata_sphinx_theme"
# html_logo = "_static/gate_logo.png"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "show_toc_level": 2,
    "use_edit_page_button": True,
}

html_context = {
    # "github_url": "https://github.com", # or your GitHub Enterprise site
    "github_user": "OpenGATE",
    "github_repo": "opengate",
    "github_version": "master",
    "doc_path": "docs/source",
}

# html_theme_options = {
#     "toc_depth": 3,
#     "navigation_depth": 2,  # Shows three levels of TOC in the sidebar
# }

# html_css_files = [
#    'css/rtd_dark.css',
# ]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}

# Configure viewcode extension.
code_url = f"https://github.com/OpenGATE/opengate/blob/master/"


def linkcode_resolve(domain, info):
    if domain != "py":
        return None
    if not info["module"]:
        return None
    filename = info["module"].replace(".", "/")
    if filename.startswith("opengate_core"):
        filename = "opengate_core/" + filename
    if os.path.isdir("../../" + filename):
        filename = filename + "/__init__"
    return code_url + filename + ".py"


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "OPENGATEdoc"

# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "OPENGATE.tex",
        "OPENGATE Documentation",
        "OpenGATE Collaboration",
        "manual",
    ),
]

# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "OPENGATE", "OPENGATE Documentation", [author], 1)]

# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "OPENGATE",
        "OPENGATE Documentation",
        author,
        "OPENGATE",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]

# -- Extension configuration -------------------------------------------------
# sphinx-autoapi configuration
# autoapi_type = "python"
# autoapi_dirs = ["../../opengate", "../../core"]
# autoapi_ignore = ["*/opengate/tests/src*"]
